@@ -27,8 +27,10 @@
 #include "config.h"
 #endif
 
+#include "gstopenh264elements.h"
 #include "gstopenh264enc.h"
 
+#include <gst/pbutils/pbutils.h>
 #include <gst/gst.h>
 #include <gst/base/base.h>
 #include <gst/video/video.h>
@@ -155,13 +157,13 @@ static gboolean gst_openh264enc_set_format (GstVideoEncoder * encoder,
     GstVideoCodecState * state);
 static GstFlowReturn gst_openh264enc_handle_frame (GstVideoEncoder * encoder,
     GstVideoCodecFrame * frame);
-static GstFlowReturn gst_openh264enc_finish (GstVideoEncoder * encoder);
 static gboolean gst_openh264enc_propose_allocation (GstVideoEncoder * encoder,
     GstQuery * query);
 static void gst_openh264enc_set_usage_type (GstOpenh264Enc * openh264enc,
     gint usage_type);
 static void gst_openh264enc_set_rate_control (GstOpenh264Enc * openh264enc,
     gint rc_mode);
+static gboolean openh264enc_element_init (GstPlugin * plugin);
 
 
 #define DEFAULT_BITRATE            (128000)
@@ -222,16 +224,17 @@ GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS
-    ("video/x-h264, stream-format=(string)\"byte-stream\", alignment=(string)\"au\", profile=(string)\"baseline\"")
-    );
+    ("video/x-h264, "
+      "stream-format=(string)\"byte-stream\", alignment=(string)\"au\","
+      "profile = (string) { constrained-baseline, baseline, main, constrained-high, high }"
+    )
+);
 
 /* class initialization */
 
 G_DEFINE_TYPE_WITH_CODE (GstOpenh264Enc, gst_openh264enc,
-    GST_TYPE_VIDEO_ENCODER,
-    G_IMPLEMENT_INTERFACE (GST_TYPE_PRESET, NULL);
-    GST_DEBUG_CATEGORY_INIT (gst_openh264enc_debug_category, "openh264enc", 0,
-        "debug category for openh264enc element"));
+    GST_TYPE_VIDEO_ENCODER, G_IMPLEMENT_INTERFACE (GST_TYPE_PRESET, NULL));
+GST_ELEMENT_REGISTER_DEFINE_CUSTOM (openh264enc, openh264enc_element_init);
 
 static void
 gst_openh264enc_class_init (GstOpenh264EncClass * klass)
@@ -261,7 +264,6 @@ gst_openh264enc_class_init (GstOpenh264EncClass * klass)
       GST_DEBUG_FUNCPTR (gst_openh264enc_handle_frame);
   video_encoder_class->propose_allocation =
       GST_DEBUG_FUNCPTR (gst_openh264enc_propose_allocation);
-  video_encoder_class->finish = GST_DEBUG_FUNCPTR (gst_openh264enc_finish);
 
   /* define properties */
   g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_USAGE_TYPE,
@@ -402,6 +404,7 @@ gst_openh264enc_init (GstOpenh264Enc * openh264enc)
   openh264enc->complexity = DEFAULT_COMPLEXITY;
   openh264enc->bitrate_changed = FALSE;
   openh264enc->max_bitrate_changed = FALSE;
+  openh264enc->profile = PRO_BASELINE;
   gst_openh264enc_set_usage_type (openh264enc, CAMERA_VIDEO_REAL_TIME);
   gst_openh264enc_set_rate_control (openh264enc, RC_QUALITY_MODE);
 }
@@ -678,6 +681,43 @@ gst_openh264enc_stop (GstVideoEncoder * encoder)
   return TRUE;
 }
 
+static guint8
+gst_openh264enc_get_level_from_caps (GstCaps *outcaps, GstCaps *allowed_caps)
+{
+  GstStructure *s = gst_caps_get_structure (outcaps, 0);
+  const gchar * level = gst_structure_get_string (gst_caps_get_structure (allowed_caps, 0), "level");
+
+  if (!level)
+    return LEVEL_UNKNOWN;
+
+  gst_structure_set (s, "level", G_TYPE_STRING, level, NULL);
+  return gst_codec_utils_h264_get_level_idc (level);
+}
+
+static EProfileIdc
+gst_openh264enc_get_profile_from_caps (GstCaps *outcaps, GstCaps *allowed_caps)
+{
+  EProfileIdc oh264_profile = PRO_BASELINE;
+  GstStructure *allowed_s = gst_caps_get_structure (allowed_caps, 0);
+  GstStructure *s = gst_caps_get_structure (outcaps, 0);
+  const gchar *profile = gst_structure_get_string (allowed_s, "profile");
+
+  if (!profile)
+    return oh264_profile;
+
+  gst_structure_set (s, "profile", G_TYPE_STRING, profile, NULL);
+  if (!g_strcmp0 (profile, "constrained-baseline") ||
+      !g_strcmp0 (profile, "baseline"))
+    return PRO_BASELINE;
+  else if (!g_strcmp0 (profile, "main"))
+    return PRO_MAIN;
+  else if (!g_strcmp0 (profile, "high") ||
+      !g_strcmp0 (profile, "constrained-high"))
+    return PRO_HIGH;
+
+  g_assert_not_reached ();
+  return PRO_BASELINE;
+}
 
 static gboolean
 gst_openh264enc_set_format (GstVideoEncoder * encoder,
@@ -694,6 +734,7 @@ gst_openh264enc_set_format (GstVideoEncoder * encoder,
   GstVideoCodecState *output_state;
   openh264enc->frame_count = 0;
   int video_format = videoFormatI420;
+  GstCaps *allowed_caps = NULL;
 
   debug_caps = gst_caps_to_string (state->caps);
   GST_DEBUG_OBJECT (openh264enc, "gst_e26d4_enc_set_format called, caps: %s",
@@ -717,10 +758,22 @@ gst_openh264enc_set_format (GstVideoEncoder * encoder,
     WelsDestroySVCEncoder (openh264enc->encoder);
     openh264enc->encoder = NULL;
   }
-  WelsCreateSVCEncoder (&openh264enc->encoder);
+
+  if (WelsCreateSVCEncoder (&openh264enc->encoder) != 0) {
+    GST_ELEMENT_ERROR (openh264enc, LIBRARY, INIT, (NULL),
+        ("Failed to create OpenH264 encoder."));
+    return FALSE;
+  }
+
   unsigned int uiTraceLevel = WELS_LOG_ERROR;
   openh264enc->encoder->SetOption (ENCODER_OPTION_TRACE_LEVEL, &uiTraceLevel);
 
+  outcaps = gst_static_pad_template_get_caps (&gst_openh264enc_src_template);
+  outcaps = gst_caps_make_writable (outcaps);
+  allowed_caps = gst_pad_get_allowed_caps (GST_VIDEO_ENCODER_SRC_PAD (encoder));
+  allowed_caps = gst_caps_make_writable (allowed_caps);
+  allowed_caps = gst_caps_fixate (allowed_caps);
+
   GST_OBJECT_LOCK (openh264enc);
 
   openh264enc->encoder->GetDefaultParams (&enc_params);
@@ -753,13 +806,17 @@ gst_openh264enc_set_format (GstVideoEncoder * encoder,
   enc_params.bPrefixNalAddingCtrl = 0;
   enc_params.fMaxFrameRate = fps_n * 1.0 / fps_d;
   enc_params.iLoopFilterDisableIdc = openh264enc->deblocking_mode;
-  enc_params.sSpatialLayers[0].uiProfileIdc = PRO_BASELINE;
+  openh264enc->profile = gst_openh264enc_get_profile_from_caps (outcaps, allowed_caps);
+  enc_params.sSpatialLayers[0].uiProfileIdc = openh264enc->profile;
+  enc_params.sSpatialLayers[0].uiLevelIdc = (ELevelIdc) gst_openh264enc_get_level_from_caps (outcaps, allowed_caps);
   enc_params.sSpatialLayers[0].iVideoWidth = enc_params.iPicWidth;
   enc_params.sSpatialLayers[0].iVideoHeight = enc_params.iPicHeight;
   enc_params.sSpatialLayers[0].fFrameRate = fps_n * 1.0 / fps_d;
   enc_params.sSpatialLayers[0].iSpatialBitrate = enc_params.iTargetBitrate;
   enc_params.sSpatialLayers[0].iMaxSpatialBitrate = enc_params.iMaxBitrate;
 
+  gst_clear_caps (&allowed_caps);
+
   if (openh264enc->slice_mode == GST_OPENH264_SLICE_MODE_N_SLICES) {
     if (openh264enc->num_slices == 1)
       slice_mode = SM_SINGLE_SLICE;
@@ -803,10 +860,6 @@ gst_openh264enc_set_format (GstVideoEncoder * encoder,
 
   openh264enc->encoder->SetOption (ENCODER_OPTION_DATAFORMAT, &video_format);
 
-  outcaps =
-      gst_caps_copy (gst_static_pad_template_get_caps
-      (&gst_openh264enc_src_template));
-
   output_state = gst_video_encoder_set_output_state (encoder, outcaps, state);
   gst_video_codec_state_unref (output_state);
 
@@ -868,98 +921,64 @@ gst_openh264enc_handle_frame (GstVideoEncoder * encoder,
 
   GST_OBJECT_UNLOCK (openh264enc);
 
-  if (frame) {
-    src_pic = new SSourcePicture;
-
-    if (src_pic == NULL) {
-      if (frame)
-        gst_video_codec_frame_unref (frame);
-      return GST_FLOW_ERROR;
-    }
-    //fill default src_pic
-    src_pic->iColorFormat = videoFormatI420;
-    src_pic->uiTimeStamp = frame->pts / GST_MSECOND;
+  src_pic = g_new0 (SSourcePicture, 1);
+  if (src_pic == NULL) {
+    gst_video_codec_frame_unref (frame);
+    return GST_FLOW_ERROR;
   }
 
+  src_pic->iColorFormat = videoFormatI420;
+  src_pic->uiTimeStamp = frame->pts / GST_MSECOND;
+
   openh264enc->frame_count++;
-  if (frame) {
-    if (G_UNLIKELY (openh264enc->frame_count == 1)) {
-      openh264enc->time_per_frame = (GST_SECOND / openh264enc->framerate);
-      openh264enc->previous_timestamp = frame->pts;
-    } else {
-      openh264enc->time_per_frame = (guint64)
-          (openh264enc->time_per_frame * 0.8 + (frame->pts -
-              openh264enc->previous_timestamp) * 0.2);
-      openh264enc->previous_timestamp = frame->pts;
-      if (openh264enc->frame_count % 10 == 0) {
-        fps = GST_SECOND / (gdouble) openh264enc->time_per_frame;
-        openh264enc->encoder->SetOption (ENCODER_OPTION_FRAME_RATE, &fps);
-      }
+  if (G_UNLIKELY (openh264enc->frame_count == 1)) {
+    openh264enc->time_per_frame = (GST_SECOND / openh264enc->framerate);
+    openh264enc->previous_timestamp = frame->pts;
+  } else {
+    openh264enc->time_per_frame = (guint64)
+        (openh264enc->time_per_frame * 0.8 + (frame->pts -
+            openh264enc->previous_timestamp) * 0.2);
+    openh264enc->previous_timestamp = frame->pts;
+    if (openh264enc->frame_count % 10 == 0) {
+      fps = GST_SECOND / (gdouble) openh264enc->time_per_frame;
+      openh264enc->encoder->SetOption (ENCODER_OPTION_FRAME_RATE, &fps);
     }
   }
 
-  if (frame) {
-    gst_video_frame_map (&video_frame, &openh264enc->input_state->info,
-        frame->input_buffer, GST_MAP_READ);
-    src_pic->iPicWidth = GST_VIDEO_FRAME_WIDTH (&video_frame);
-    src_pic->iPicHeight = GST_VIDEO_FRAME_HEIGHT (&video_frame);
-    src_pic->iStride[0] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 0);
-    src_pic->iStride[1] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 1);
-    src_pic->iStride[2] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 2);
-    src_pic->pData[0] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 0);
-    src_pic->pData[1] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 1);
-    src_pic->pData[2] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 2);
+  gst_video_frame_map (&video_frame, &openh264enc->input_state->info,
+      frame->input_buffer, GST_MAP_READ);
+  src_pic->iPicWidth = GST_VIDEO_FRAME_WIDTH (&video_frame);
+  src_pic->iPicHeight = GST_VIDEO_FRAME_HEIGHT (&video_frame);
+  src_pic->iStride[0] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 0);
+  src_pic->iStride[1] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 1);
+  src_pic->iStride[2] = GST_VIDEO_FRAME_COMP_STRIDE (&video_frame, 2);
+  src_pic->pData[0] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 0);
+  src_pic->pData[1] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 1);
+  src_pic->pData[2] = GST_VIDEO_FRAME_COMP_DATA (&video_frame, 2);
 
-    force_keyframe = GST_VIDEO_CODEC_FRAME_IS_FORCE_KEYFRAME (frame);
-    if (force_keyframe) {
-      openh264enc->encoder->ForceIntraFrame (true);
-      GST_DEBUG_OBJECT (openh264enc,
-          "Got force key unit event, next frame coded as intra picture");
-    }
+  force_keyframe = GST_VIDEO_CODEC_FRAME_IS_FORCE_KEYFRAME (frame);
+  if (force_keyframe) {
+    openh264enc->encoder->ForceIntraFrame (true);
+    GST_DEBUG_OBJECT (openh264enc,
+        "Got force key unit event, next frame coded as intra picture");
   }
 
   memset (&frame_info, 0, sizeof (SFrameBSInfo));
   ret = openh264enc->encoder->EncodeFrame (src_pic, &frame_info);
   if (ret != cmResultSuccess) {
-    if (frame) {
-      gst_video_frame_unmap (&video_frame);
-      gst_video_codec_frame_unref (frame);
-      delete src_pic;
-      GST_ELEMENT_ERROR (openh264enc, STREAM, ENCODE,
-          ("Could not encode frame"), ("Openh264 returned %d", ret));
-      return GST_FLOW_ERROR;
-    } else {
-      return GST_FLOW_EOS;
-    }
-  }
-
-  if (videoFrameTypeSkip == frame_info.eFrameType) {
-    if (frame) {
-      gst_video_frame_unmap (&video_frame);
-      gst_video_encoder_finish_frame (encoder, frame);
-      delete src_pic;
-    }
-
-    return GST_FLOW_OK;
-  }
-
-  if (frame) {
     gst_video_frame_unmap (&video_frame);
     gst_video_codec_frame_unref (frame);
-    delete src_pic;
-    src_pic = NULL;
-    frame = NULL;
+    g_free (src_pic);
+    GST_ELEMENT_ERROR (openh264enc, STREAM, ENCODE,
+        ("Could not encode frame"), ("Openh264 returned %d", ret));
+    return GST_FLOW_ERROR;
   }
 
-  /* FIXME: openh264 has no way for us to get a connection
-   * between the input and output frames, we just have to
-   * guess based on the input */
-  frame = gst_video_encoder_get_oldest_frame (encoder);
-  if (!frame) {
-    GST_ELEMENT_ERROR (openh264enc, STREAM, ENCODE,
-        ("Could not encode frame"), ("openh264enc returned %d", ret));
-    gst_video_codec_frame_unref (frame);
-    return GST_FLOW_ERROR;
+  gst_video_frame_unmap (&video_frame);
+  g_free (src_pic);
+
+  if (videoFrameTypeSkip == frame_info.eFrameType) {
+    return gst_video_encoder_finish_frame (encoder, frame);
   }
 
   if (videoFrameTypeIDR == frame_info.eFrameType) {
@@ -977,33 +996,28 @@ gst_openh264enc_handle_frame (GstVideoEncoder * encoder,
   frame->output_buffer =
       gst_video_encoder_allocate_output_buffer (encoder, buf_length);
 
-
   buf_length = 0;
   for (i = 0; i < frame_info.iLayerNum; i++) {
     gsize layer_size = 0;
     for (j = 0; j < frame_info.sLayerInfo[i].iNalCount; j++) {
       layer_size += frame_info.sLayerInfo[i].pNalLengthInByte[j];
     }
-    gst_buffer_fill (frame->output_buffer, buf_length, frame_info.sLayerInfo[i].pBsBuf, layer_size);
+    gst_buffer_fill (frame->output_buffer, buf_length,
+        frame_info.sLayerInfo[i].pBsBuf, layer_size);
     buf_length += layer_size;
   }
 
-  GST_LOG_OBJECT (openh264enc, "openh264 picture %scoded OK!",
-      (ret != cmResultSuccess) ? "NOT " : "");
+  GST_LOG_OBJECT (openh264enc, "openh264 picture coded OK!");
 
   return gst_video_encoder_finish_frame (encoder, frame);
 }
 
-static GstFlowReturn
-gst_openh264enc_finish (GstVideoEncoder * encoder)
+static gboolean
+openh264enc_element_init (GstPlugin * plugin)
 {
-  GstOpenh264Enc *openh264enc = GST_OPENH264ENC (encoder);
+  GST_DEBUG_CATEGORY_INIT (gst_openh264enc_debug_category, "openh264enc", 0,
+      "debug category for openh264enc element");
 
-  if (openh264enc->frame_count == 0)
-    return GST_FLOW_OK;
-
-  /* Drain encoder */
-  while ((gst_openh264enc_handle_frame (encoder, NULL)) == GST_FLOW_OK);
-
-  return GST_FLOW_OK;
+  return gst_element_register (plugin, "openh264enc", GST_RANK_MARGINAL,
+      GST_TYPE_OPENH264ENC);
 }
