@@ -561,6 +561,7 @@ gst_rist_sink_init (GstRistSink * sink)
   GstPadTemplate *pad_template;
 
   sink->rtpext = gst_element_factory_make ("ristrtpext", "ristrtpext");
+  gst_object_ref_sink (sink->rtpext);
 
   g_mutex_init (&sink->bonds_lock);
   sink->bonds = g_ptr_array_new ();
@@ -1322,7 +1323,8 @@ gst_rist_sink_finalize (GObject * object)
   }
   g_ptr_array_free (sink->bonds, TRUE);
 
-  g_clear_object (&sink->rtxbin);
+  gst_clear_object (&sink->rtxbin);
+  gst_clear_object (&sink->rtpext);
 
   g_mutex_unlock (&sink->bonds_lock);
   g_mutex_clear (&sink->bonds_lock);
@@ -1339,10 +1341,10 @@ gst_rist_sink_class_init (GstRistSinkClass * klass)
 
   session_id_quark = g_quark_from_static_string ("gst-rist-sink-session-id");
 
-  gst_element_class_set_metadata (element_class,
+  gst_element_class_set_static_metadata (element_class,
       "RIST Sink", "Source/Network",
       "Sink that implements RIST TR-06-1 streaming specification",
-      "Nicolas Dufresne <nicolas.dufresne@collabora.com");
+      "Nicolas Dufresne <nicolas.dufresne@collabora.com>");
   gst_element_class_add_static_pad_template (element_class, &sink_templ);
 
   element_class->change_state = gst_rist_sink_change_state;
